%   MATLAB script: 
%   Four point algorithm for decomposition of homography
%   June '08, CVPR 2008 Tutorial on Computational Symmetry
%
%   Author: Allen Y. Yang - yang@eecs.berkeley.edu
%
%   Reference: An Invitation to 3-D Vision, MaSKS, Springer.
%
%   This package is free to use without restrictions. No support is
%   guaranteed.
close all
clear

addpath helper_functions

% Step 1: generate a planar structure
noiseLevel = 0;
sampleCount = 50;
outlierPercentage = 0;
[img1, img2, aprioriLabels, aprioriMotions] = generate_scene(2, sampleCount, noiseLevel, outlierPercentage,2);
aprioriMotion = aprioriMotions{1};  % We only use one planar object
aprioriH = aprioriMotion.R + aprioriMotion.T*aprioriMotion.N.'/aprioriMotion.d;

% Step 2: estimate homography matrix H
img1 = [img1; ones(1,sampleCount)];
img2 = [img2; ones(1,sampleCount)];
A = zeros(3*sampleCount,9);
for sampleIndex=1:sampleCount
    A((sampleIndex-1)*3+1:sampleIndex*3,:) = kron(img1(:,sampleIndex),hat(img2(:,sampleIndex)))';
end
[U, S, V] = svd(A);
Hstacked = V(:,end);
H = reshape(Hstacked,[3 3]);

% Step 3: four-point algorithm
% 3.1. normalize the second singular value
[U, S, V] = svd(H);
H = H/S(2,2);

% 3.2. normalize the sign of matrix H
signArray = zeros(1,sampleCount);
for sampleIndex=1:sampleCount
    signArray(sampleIndex) = sign(img2(:,sampleIndex).'*H*img1(:,sampleIndex));
end
if any(signArray~=signArray(1))
    warning('Tested pairs do not have uniform depth sign!');
end
if sum(signArray)<0
    H = -H;
end

% Verify the accuracy of H
currentResidual = norm(H-aprioriH);
disp(['The matrix norm between estimated H and apriori H is ' num2str(currentResidual) '.']);

if currentResidual>0.1
    warning('The object positions are degenerate! ');
end

% 3.3. compute R, t, N
solutions= cell(1,4);
[U, S, V] = svd(H.'*H);
v1 = V(:,1)/norm(V(:,1)); v2 = V(:,2)/norm(V(:,2)); v3 = V(:,3)/norm(V(:,3));
u1 = (sqrt(1-S(3,3))*v1+sqrt(S(1,1)-1)*v3)/sqrt(S(1,1)-S(3,3));
u2 = (sqrt(1-S(3,3))*v1-sqrt(S(1,1)-1)*v3)/sqrt(S(1,1)-S(3,3));

solutions{1}.R = [H*v2, H*u1, hat(H*v2)*H*u1]*[v2, u1, hat(v2)*u1].';
solutions{1}.N = hat(v2)*u1;
solutions{1}.t = (H-solutions{1}.R)*solutions{1}.N;
solutions{1}.t = solutions{1}.t/norm(solutions{1}.t);
solutions{2}.R = solutions{1}.R;
solutions{2}.N = -solutions{1}.N;
solutions{2}.t = -solutions{1}.t;
solutions{3}.R = [H*v2, H*u2, hat(H*v2)*H*u2]*[v2, u2, hat(v2)*u2].';
solutions{3}.N = hat(v2)*u2;
solutions{3}.t = (H-solutions{3}.R)*solutions{3}.N;
solutions{3}.t = solutions{3}.t/norm(solutions{3}.t);
solutions{4}.R = solutions{3}.R;
solutions{4}.N = -solutions{3}.N;
solutions{4}.t = -solutions{3}.t;

% Step 4. reject 2 solutions by positive depth constraint
for solutionIndex=1:4
    if solutions{solutionIndex}.N(3)>=0
        disp(['Solution #' num2str(solutionIndex) ' is a possible solution.']);
    end
end

% Finally, reveal the true solution using the prior information
minMatrixNorm = inf;
aprioriMotion.T = aprioriMotion.T/norm(aprioriMotion.T);
for solutionIndex=1:4
    currentNorm = norm(aprioriMotion.R-solutions{solutionIndex}.R) + ...
        norm(aprioriMotion.T - solutions{solutionIndex}.t);
    if currentNorm<minMatrixNorm
        minMatrixNorm = currentNorm;
        trueSolutionIndex = solutionIndex;
    end
end

disp(['The true solution is #' num2str(trueSolutionIndex) '.']);