function [firstImage, secondImage, trueLabels, trueMotions] = generate_scene(motionTypes, sampleCounts, noiseLevel, outlierPercentage, DEBUG)

% generate_scene
%
%   Generate feature correspondences from two perspective views of a
%   dynamic scene with multiple rigid bodys and planar structures.
%
%   Inputs:
%
%       motionTypes - A list of objects, specified by type. Valid entries in the list
%               are:  1 - a general rigid body (3-D point cloud with radius
%               4), 2 - a planar structure (a square plane with edges of
%               length 6).
%
%       sampleCounts - A list of sample counts for each of the objects in
%               scene
%
%       noiseLevel - the amount of uniformed noise to add to the scene. The
%       unit is pixel.
%                       
%       outlierPercentage - the percentage of outliers to add to the scene.
%               the outliers are uniformly distributed in the scene.
%
%
%   Outputs:
%       
%       firstImage - the 2-D coordinates of each feature in the first image
%               of the scene.
%
%       secondImage - the 2-D coordinates of each feature in the second
%               image of the scene.
%
%       trueLabels - the true group labels for each feature correspondence. the label -1
%               designates a sample as an outlier.
%
%       trueMotions - the true 3-D motions from the object(s) to the two
%               camera centers.
%
%   June '08
%       Questions? Allen Y. Yang - yang@eecs.berkeley.edu
%
%   This package is free to use without restrictions. No support is
%   guaranteed.
%   


if nargin < 3
    noiseLevel = 0;
end

if nargin < 4
    outlierPercentage = 0;
end

if nargin<5
    DEBUG = 0;
end

RIGID_BODY_MOTION = 1;
PLANAR_HOMOGRAPHY = 2;

normalizedImageSize=1024;

objectCount = length(motionTypes);
totalSampleCount = sum(sampleCounts);

sampleIndex = 0;
data3d = zeros(3, totalSampleCount);
scene1 = data3d;
scene2 = data3d;
trueLabels = zeros(1, totalSampleCount);
trueMotions = cell(1,objectCount);

% First, generate each object in the 3d world at canonical positions
for objectIndex = 1:objectCount
    
    trueLabels(sampleIndex+1:sampleIndex+sampleCounts(objectIndex)) = objectIndex;
    % A rigid object is a point cloud with radius 4
    if motionTypes(objectIndex) == RIGID_BODY_MOTION        
        radius = 4;
        for currentSampleIndex=1:sampleCounts(objectIndex)            
            b =  randn(3,1);
            sampleIndex = sampleIndex + 1;
            data3d(:, sampleIndex) = radius * (b ./ norm(b)) ;            
        end
    end
    
    % A plane will be a flat with edges of length 12
    if motionTypes(objectIndex) == PLANAR_HOMOGRAPHY
        edgeLength = 4;
        for currentSampleIndex=1:sampleCounts(objectIndex)
            %theta = rand*(2*pi);
            %rr = rand*radius;
            x = (rand - 0.5) * (2*edgeLength);
            y = (rand - 0.5) * (2*edgeLength);
            sampleIndex = sampleIndex + 1;
            %data3d(:, sampleIndex) = [rr*cos(theta); rr*sin(theta); 0];
            data3d(:, sampleIndex) = [x ; y ; 0];
        end
    end
    
end

% Generate to camera coordinate transform
panCamera = pi/10;
Rcamera = [cos(panCamera) 0 -sin(panCamera); 0 1 0; sin(panCamera) 0 cos(panCamera)];
Tcamera = [0; 0; 30];

% 
% Create the first scene
sampleStartIndex=0;
for objectIndex=1:objectCount
    rotationAngle = pi/16*randn();
    R = [cos(rotationAngle), 0, -sin(rotationAngle);0 1 0; sin(rotationAngle), 0, cos(rotationAngle)];
    rotationAngle = pi/16*randn();
    R = R*[cos(rotationAngle), -sin(rotationAngle), 0; sin(rotationAngle), cos(rotationAngle), 0; 0, 0, 1];
    T = [80;80;20].*rand(3,1)-[40;40;0];
    sampleIndices = sampleStartIndex+1:sampleStartIndex+sampleCounts(objectIndex);
    scene1(:, sampleIndices) = R*data3d(:, sampleIndices) + repmat(T, 1, sampleCounts(objectIndex));
    sampleStartIndex = sampleStartIndex + sampleCounts(objectIndex);
    
    % Register the motion
    R = Rcamera*R;
    T = Rcamera*T + Tcamera;
    trueMotions{objectIndex}.R0 = R;
    trueMotions{objectIndex}.T0 = T;
    if motionTypes(objectIndex) == PLANAR_HOMOGRAPHY
        trueMotions{objectIndex}.N = R*[0;0;1];
        trueMotions{objectIndex}.d = abs(T.'*trueMotions{objectIndex}.N);
    end
end

% Create the second scene
sampleStartIndex=0;
for objectIndex=1:objectCount
    rotationAngle = pi/16*randn();
    R = [cos(rotationAngle), 0, -sin(rotationAngle);0 1 0; sin(rotationAngle), 0, cos(rotationAngle)];
    rotationAngle = pi/16*randn();
    R = R*[cos(rotationAngle), -sin(rotationAngle), 0; sin(rotationAngle), cos(rotationAngle), 0; 0, 0, 1];
    T = [80;80;20].*rand(3,1)-[40;40;0];
    sampleIndices = sampleStartIndex+1:sampleStartIndex+sampleCounts(objectIndex);
    scene2(:, sampleIndices) = R*data3d(:, sampleIndices) + repmat(T, 1, sampleCounts(objectIndex));
    sampleStartIndex = sampleStartIndex + sampleCounts(objectIndex);
    
    % Register the motion
    R = Rcamera*R;
    T = Rcamera*T + Tcamera;
    trueMotions{objectIndex}.R = R*inv(trueMotions{objectIndex}.R0);
    trueMotions{objectIndex}.T = -trueMotions{objectIndex}.R*trueMotions{objectIndex}.T0+T;
end

scene1Camera = Rcamera*scene1 + repmat(Tcamera, 1, totalSampleCount);
scene2Camera = Rcamera*scene2 + repmat(Tcamera, 1, totalSampleCount);

% Render the scene
if DEBUG >0
    % render the scene
    figure; 
    subplot(1,2,1); hold on
    plot3(scene1Camera(1,:), scene1Camera(2,:), scene1Camera(3,:), 'k*', 'MarkerSize',5);
    v0 = [0;0;0];
    vx = [3;0;0];
    vy = [0;3;0];
    vz = [0;0;5];
    arrow3(v0',vx','1k',1,2);
    arrow3(v0',vy','1k',1,2);
    arrow3(v0',vz','1.5r',2,2);      
    axis equal
    
    
    subplot(1,2,2); hold on
    plot3(scene2Camera(1,:), scene2Camera(2,:), scene2Camera(3,:), 'k*', 'MarkerSize',5);
    v0 = [0;0;0];
    vx = [3;0;0];
    vy = [0;3;0];
    vz = [0;0;5];
    arrow3(v0',vx','1k',1,2);
    arrow3(v0',vy','1k',1,2);
    arrow3(v0',vz','1.5r',2,2);      
    axis equal
end

% Perspective projection assuming camera is fixed
firstImage = scene1Camera(1:2,:) ./ [scene1Camera(3,:); scene1Camera(3,:)];
secondImage = scene2Camera(1:2, :) ./ [scene2Camera(3,:); scene2Camera(3,:)];

firstImageLowerBounds = min(firstImage, [], 2);
firstImageUpperBounds = max(firstImage, [], 2);
secondImageLowerBounds = min(secondImage, [], 2);
secondImageUpperBounds = max(secondImage, [], 2);

if (outlierPercentage > 0)
    outlierCount = floor((outlierPercentage * totalSampleCount)/(1 - outlierPercentage));

    firstImageOutliers = rand(2,outlierCount).*repmat(firstImageUpperBounds - firstImageLowerBounds, 1, outlierCount) + repmat(firstImageLowerBounds, 1, outlierCount);
    secondImageOutliers = rand(2,outlierCount).*repmat(secondImageUpperBounds - secondImageLowerBounds, 1, outlierCount) + repmat(secondImageLowerBounds, 1, outlierCount);

    firstImage = [firstImage firstImageOutliers];
    secondImage = [secondImage secondImageOutliers];
    trueLabels = [trueLabels repmat(-1, 1, outlierCount)];
    totalSampleCount = totalSampleCount + outlierCount;
end

% compute rescale factor
maxPixelNorm = max(norm(firstImageLowerBounds-firstImageUpperBounds),norm(secondImageLowerBounds-secondImageUpperBounds));
firstImage = firstImage + noiseLevel/normalizedImageSize*maxPixelNorm*2*(rand(2, totalSampleCount)-0.5*ones(2,totalSampleCount));
secondImage = secondImage + noiseLevel/normalizedImageSize*maxPixelNorm*2*(rand(2, totalSampleCount)-0.5*ones(2,totalSampleCount));

% Render the scene
if DEBUG >1
    % render the scene
    figure; 
    subplot(1,2,1); 
    plot(firstImage(1,:), firstImage(2,:), 'k*', 'MarkerSize',5);
    axis equal
    
    subplot(1,2,2); 
    plot(secondImage(1,:), secondImage(2,:), 'k*', 'MarkerSize',5);
    axis equal

end